//
// tngBot.cpp
//
// Last Revised: 22 July 1998
//

#include <stdio.h>
#include <memory.h>
#include "tngbot.h"

TraderInfo::TraderInfo (void)
{
	trader 		= 0;
	lastPayoff	= 0.0f;
	sumPayoffs	= 0.0f;
	numPayoffs	= 0;
	expPayoff 	= 0.0f;
	numTrades 	= 0;
	offerStatus = NO_OFFER;
}

static int
compTraderInfo (const void* a, const void* b)
//: Compare two TraderInfo instance, sorting on expected payoff.
{
	// sort according to payoff in descending order
	TraderInfo*	p1 = (TraderInfo*) a;
	TraderInfo*	p2 = (TraderInfo*) b;

	int			comp = CompareFloats(p1->expPayoff, p2->expPayoff);

	//printf ("compTraderInfo(%.4f, %.4f)->%d\n", p1->expPayoff, p2->expPayoff, comp);

	return comp;
}

//////////////////////////
//
// tngTradeBot

int					tngTradeBot::_Memory = 1;
int					tngTradeBot::_StateBits = 2;

tngTradeBot::tngTradeBot (bioGType* dna) : bioPType(dna)
//: TradeBot constructor -- initializes member data with default values.
{
	_sim = 0;

	_maxOffers 	= 0;
	_maxAccepts = 0;

	_sumTotalPayoffs = 0.0f;
	_numTotalPayoffs = 0;
	_offers 	= 0;
	_offersMade = 0;
	_accepts 	= 0;

	_traderCount = 0;
	_traderList	= 0;
	_waitList = 0;

	_action = 1;

	InitProgram();
}

tngTradeBot::~tngTradeBot (void)
//: Clean up arrays we allocated.
{
//	printf ("dtor %p\n", this);

	delete [] _traderList;
	delete [] _waitList;
}

void
tngTradeBot::SetMemory (int m)
{
	ASSERT (!(m < 0));

	_Memory = m;
}

void
tngTradeBot::SetStates (int s)
{
	ASSERT (s > 0);

	_StateBits = (int) ceil(log((double)s)/log(2.0));
}

void
tngTradeBot::SetStateBits (int b)
{
	ASSERT (b > 0);

	_StateBits = b;
}

int
tngTradeBot::FSMSize (void)
{
	return (1 << (_Memory + _StateBits));
}

void
tngTradeBot::InitProgram (void)
//: Creates a finite state machine (an instance of bioFSM)
//: to use as our program for determing trade actions.
{
	int             size = FSMSize(); 
	int             len;
	int*            array = new int[size]; ASSERT (array);
	bioFSM*	        rule;

	len = _gtype->GetLength()/8;  // get number of bytes in gtype
	ASSERT (len == size);

	// Extract an array of integers from the genotype to initialize
	// the FSM
	for (int i = 0; i < len; i++)
	{
		array[i] = (int) ((bioHapGType*)_gtype)->GetByte(i);
		//printf("%d %d\n", i, array[i]);
	}

	// Create a new instance of FSM with _Memory input bits,
	// _StateBits state bits, and one output bit. The input bits are the
	// last trade actions the other tradebot made with us. The one
	// output bit is interpreted as 0 = defect, 1 = cooperate.
	rule = new bioFSM (_Memory, _StateBits, 1);
	ASSERT_VALID (rule);

	rule->Init();
	rule->ReadArray (array, len);

	SetProgram (rule);
	delete array;
}

void
tngTradeBot::Init (tngSimulation* sim, bioList& traders, int maxOffers,
					int maxAccepts)
{
	tngTradeBot*			tradebot;
	int						i;

	_sim = sim;
	_maxOffers = maxOffers;
	_maxAccepts = maxAccepts;

	// Allocate an array of information about the other traders in
	// the game.
	_traderCount = traders.Count();
	_traderList = new TraderInfo[_traderCount]; 
	ASSERT (_traderList != 0);

	if (IsSeller())
	{
		_waitList = new TraderInfo[_traderCount];
		ASSERT (_waitList != 0);
	}

	for (i = 0, traders.First(); tradebot =
			(tngTradeBot*) traders.Next(); i++)
	{
		_traderList[i].trader = tradebot;

		// If the tradebot is me, the initial payoff is set to -1 so
		// I don't try to trade with myself, otherwise set to initial
		// expected payoff.
		_traderList[i].expPayoff = (tradebot == this) ? -1.0f :
			_sim->InitExpPayoff();
	}  
	
	if (IsSeller())
	{
		if (IsBuyer())
		{
			sprintf (_name, "[BuyerSeller %d]", _id + 1);
		}
		else
		{
			sprintf (_name, "[PureSeller %d]", _id + 1);
		}
	}
	else
	{
		sprintf (_name, "[PureBuyer %d]", _id + 1);
	}
}

void
tngTradeBot::PrepareOffers (void)
{
	for (int i = 0; i < _traderCount; i++)
	{
		_traderList[i].offerStatus = TraderInfo::NO_OFFER;
	}

	_offers = 0;
	_offersMade = 0;
}

// Print contents of TraderInfo array for debug purposes
static void
DumpList(TraderInfo* list, int count)
{
	for (int i = 0; i < count; i++)
	{
		printf ("    list-> %3d %20s %f\n", i, list[i].trader->GetName(), list[i].expPayoff);
	}
}

int
tngTradeBot::SubmitOffers (void)
//: Submit offers to the best sellers. (This message is sent only
//: to buyers.)
{
	int				i;
	int				newOffers = 0;

	//printf("begin SubmitOffers %s\n", GetName());

	if (_offers == _maxOffers)
	{
		return 0;
	}

	// Sort the seller information based on expected payoff.
	qsort (_traderList, _traderCount, sizeof (TraderInfo),
			compTraderInfo);
	//DumpList(_traderList, _traderCount);

	// Submit offers up to my quota.
	for (i = 0; i < _traderCount; i++)
	{
		// If a seller is not tolerable, quit submitting. (All sellers
		// after this one will also be intolerable because the
		// information is sorted.)
		if (CompareFloats(_traderList[i].expPayoff, 0.0f) == 1)	// !!! changed to CompareFloats dbm 97.02.22
		{
			break;
		}

		// If we haven't already been rejected or accepted, make
		// an offer...               
		ASSERT (_traderList[i].trader);
		
		if (_traderList[i].trader->IsSeller() &&
			_traderList[i].offerStatus == TraderInfo::NO_OFFER)
		{
			//printf("%s is making an offer to %s %f\n", GetName(), _traderList[i].trader->GetName(), _traderList[i].expPayoff);

			_traderList[i].trader->TakeOffer (this);
			_traderList[i].offerStatus = TraderInfo::OFFER_MADE;
			++_offers;
			++_offersMade;
			++newOffers;
		}

		// If we reached our maximum, quit now.
		if (_offers == _maxOffers)
		{
			break;
		}
	}

	//printf("%s submitted %d new offers\n", GetName(), newOffers);

	return newOffers;
}

void
tngTradeBot::TakeOffer (tngTradeBot* buyer)
//: Take an offer from a buyer and add it to my waiting list.
//: (This message is sent only to sellers.)
{
	ASSERT (_accepts < _traderCount);

	TraderInfo*			info = GetTraderInfo (buyer);
	ASSERT (info != 0);

	_waitList[_accepts].trader = buyer;
	_waitList[_accepts].expPayoff = info->expPayoff;
	
	//printf("%s got an offer from %s %f\n", GetName(), _waitList[_accepts].trader->GetName(), _waitList[_accepts].expPayoff);

	_accepts++;
}

void
tngTradeBot::WallflowerCheck (void)
{
	if (_offersMade == 0 && _accepts == 0)
	{
		_sumTotalPayoffs += _sim->WallflowerPayoff();
		++_numTotalPayoffs;
	}
}

void
tngTradeBot::AcceptOffers (void)
//: Accept offers from the best buyers. (This message will be sent
//: only to sellers.)
{
	int					i;
	int					max;
	int					rejections = 0;

	//printf("begin AcceptOffers %s\n", GetName());

	// Sort the waiting list based on expected payoff.
	qsort (_waitList, _accepts, sizeof (TraderInfo), compTraderInfo);
	//DumpList(_waitList, _accepts);

	// If we got more offers than our quota we will have to refuse some
	// (potentially) tolerable buyers.
	max = (_accepts > _maxAccepts) ? _maxAccepts : _accepts;

	for (i = 0; i < max; i++)
	{
		// If this guy is not tolerable, quit accepting offers.
		if (CompareFloats(_waitList[i].expPayoff, 0.0f) == 1) break; // !!! changed dbm 97.02.22

		//printf("  accepted %s %f\n", _waitList[i].trader->GetName(), _waitList[i].expPayoff);
	}

	//printf ("%s accepted %d offers\n", GetName(), i);

	// Refuse the rest.
	for ( ; i < _accepts; i++)
	{ 
		ASSERT (_waitList[i].trader);

		//printf("  rejecting %s %f\n", _waitList[i].trader->GetName(), _waitList[i].expPayoff);

		_waitList[i].trader->OfferRejected (this);
		++rejections;
	}

	_accepts -= rejections;
}

void
tngTradeBot::PlayPD (void)
//: Trade with the best buyers. (This message will be sent only
//: to sellers.)
{
	ASSERT_VALID (_sim);

	for (int i = 0; i < _accepts; i++)
	{
		// play PD with other
		_sim->MediateTrade (this, _waitList[i].trader);
	}

	_accepts = 0;
}

int
tngTradeBot::GetAction (tngTradeBot* bot)
// Cooperate or defect with bot?
{
	TraderInfo*		info = GetTraderInfo (bot);
	const int		MAX = 32;
	int				actions[MAX];
	int				action;

	ASSERT (_Memory < MAX);

	for (int i = 0; i < _Memory; i++)
	{
		int			idx = info->numTrades - i - 1;
		
		if (idx >= 0)
		{
			actions[i] = (int)info->tradeHistory[idx];
		}
		else
		{
			actions[i] = 1; // co-operate
		}

		//printf("GetAction actions[%d] = %d\n", i, actions[i]);
	}

	_program->SetState(info->state);

	action = _program->Think (actions, _Memory);

	_program->GetState(info->state);

	return action;
}

void
tngTradeBot::OfferRejected (tngTradeBot* seller) 
//: Receive a refusal :-(
{
	//printf ("%s got a refusal from %s\n", GetName(), seller->GetName());

	TraderInfo*		info = GetTraderInfo (seller);
	ASSERT (info != 0);

    float			payoff = _sim->RefusalPayoff();

	--_offers;
	_sumTotalPayoffs += payoff;
	++_numTotalPayoffs; 	

	info->lastPayoff = payoff;
	info->sumPayoffs += payoff;
	++info->numPayoffs;
	info->offerStatus = TraderInfo::OFFER_REJECTED;

    UpdateExp (info);
}

void
tngTradeBot::Buy (tngTradeBot* seller, float payoff, int action)
//: Update information about a trade.
{
	TraderInfo*		info = GetTraderInfo (seller);
	ASSERT (info != 0);
	                                                 
	_sumTotalPayoffs += payoff; 
    ++_numTotalPayoffs;

	info->lastPayoff = payoff;
	info->sumPayoffs += payoff;
	++info->numPayoffs;
	++info->numTrades;
	info->tradeHistory[info->numTrades-1] = action;

	UpdateExp (info);
}

void
tngTradeBot::Sell (tngTradeBot* buyer, float payoff, int action)
//: Update information about a trade.
{
	TraderInfo*		info = GetTraderInfo (buyer);
	ASSERT (info != 0);

	_sumTotalPayoffs += payoff;
	++_numTotalPayoffs;

	info->lastPayoff = payoff;
	info->sumPayoffs += payoff;
	++info->numPayoffs;
	++info->numTrades;
	info->tradeHistory[info->numTrades-1] = action;

	UpdateExp (info);
}

void
tngTradeBot::UpdateExp (TraderInfo* info)
{
	float			memWeight;

	memWeight = (float) info->numPayoffs/(info->numPayoffs + 1);

	info->expPayoff = (memWeight * info->expPayoff) +
		(1.0f - memWeight) * info->lastPayoff; 

	//printf("UpdateExp %s %f\n", info->trader->GetName(), info->expPayoff);
}

float
tngTradeBot::CalcFitness (void)
//: The fitness is 2 raised to the power of the fitness score for use
//: in the GA.  The fitness has to be a positive for the GA to work 
//: (as implemented) even if the fitness score is negative.
{
	ASSERT(_numTotalPayoffs > 0);

	float		fitnessScore = (float)_sumTotalPayoffs/_numTotalPayoffs;

	_fitness = pow (2.0, (double)fitnessScore);
	 
	return fitnessScore;
}

TraderInfo*
tngTradeBot::GetTraderInfo (tngTradeBot* bot)
//: Look up information about the tradebot in the list.
{
	for (int i = 0; i < _traderCount; i++)
	{
		if (_traderList[i].trader == bot)
		{
			return &_traderList[i];
		}
	}

	return 0;
}

void
tngTradeBot::Dump (void)
//: Print information about this tradebot.
{
	TraderInfo*		list = _traderList;
	int		   		numTraders = _traderCount;
	float			fitnessScore = CalcFitness();

  qsort (list, numTraders, sizeof (TraderInfo), compTraderInfo) ;

	printf("\n%s:  sumTotalPayoffs = %5.2f;  fitScore = %5.2f",
			GetName(), _sumTotalPayoffs, fitnessScore);

	printf("\nStored other bot info for this tradebot:\nlist#;tag;");
	printf("sumPayoffs;numPayoffs;expPayoff;numTrades;tradeHistory\n");

	for (int i = 0; i < numTraders; i++)
	{
		printf ("%3d %19s %+7.2f %3d %5.2f %3d:", i+1, list[i].trader->GetName(),
		list[i].sumPayoffs, list[i].numPayoffs, list[i].expPayoff,
		list[i].numTrades);

		for (int j = 0; j < list[i].numTrades; j++)
		{
			printf ("%s", list[i].tradeHistory[j] ? "c":"d");

			// Wrap lines to 68 characters wide.
			if (((j-21) % 68) == 0)
			{
				printf ("\n");
			}
		}
		printf ("\n");
	}
}


