//
// tngSim.cpp 
//
// Last Revised: 29 July 1998
//

#include <stdio.h>
#include <time.h>
#include "tngsim.h"

//////////////////////////
//
// tngSimulation methods

tngSimulation::tngSimulation(void)
{
}

tngSimulation::~tngSimulation(void)
{ 
	_traders.RemoveAll();
	_buyers.RemoveAll();
	_sellers.RemoveAll();

}

void
tngSimulation::Init(void)
{    		
	char*	iniFile = "tng.ini";
	float	mrate = GetInitFloat(iniFile, "MutationRate", .005f);
	int		elite = GetInitInt(iniFile, "Elite", 1);
	int		seed  = GetInitInt(iniFile, "RandomSeed", 20);

	int		fsmMemory = GetInitInt(iniFile, "FsmMemory", 1);
	int		fsmStates = GetInitInt(iniFile, "FsmStates", 16);

	srand((unsigned int) seed);

	_gmax = GetInitInt(iniFile, "GMax", 10);
	_imax = GetInitInt(iniFile, "IMax", 10);

	_traderCount     = GetInitInt(iniFile, "TraderCount", 9);
	_pureBuyerCount  = GetInitInt(iniFile, "PureBuyerCount", 3);
	_pureSellerCount = GetInitInt(iniFile, "PureSellerCount", 3);
	_buysellCount    = GetInitInt(iniFile, "BuySellCount", 3);

	ASSERT((_pureBuyerCount + _pureSellerCount + _buysellCount) == _traderCount);

	_buyerCount  = _pureBuyerCount  + _buysellCount;
	_sellerCount = _pureSellerCount + _buysellCount;

	_buyerQuota  = GetInitInt(iniFile, "BuyerQuota", 1);
	_sellerQuota = GetInitInt(iniFile, "SellerQuota", 1);

	_initExpPayoff 	  = GetInitFloat(iniFile, "InitExpPayoff", 1.0f);
	_refusalPayoff    = GetInitFloat(iniFile, "RefusalPayoff", -1.0f);
	_wallflowerPayoff = GetInitFloat(iniFile, "WallflowerPayoff", 0.0f);

	// Initialize prisoner dilemma payoffs
	_bothCoop 	= GetInitFloat(iniFile, "BothCoop", 1.0f);
	_bothDefect = GetInitFloat(iniFile, "BothDefect", -1.0f);
	_sucker 	= GetInitFloat(iniFile, "Sucker", -2.0f);
	_temptation = GetInitFloat(iniFile, "Temptation", 2.0f);

	_printGen = GetInitInt(iniFile, "PrintGen", 2);
	_printIt  = GetInitInt(iniFile, "PrintIt", 3);

	_trades = 0;

	tngTradeBot::SetMemory(fsmMemory);
	tngTradeBot::SetStates(fsmStates);

	struct tm *newtime;
	time_t aclock;

	time(&aclock); 
	newtime = localtime(&aclock); 
	printf("\nCurrent Date and Time: %s\n\n", asctime(newtime));

	printf("GMax=%d, IMax=%d\n", _gmax,  _imax );

	printf("randSeed=%d\n", seed );
	printf("mutation rate=%6.3f\n", mrate );

	printf("TraderCount: ");
	printf("Total=%d, PureBuyer=%d, PureSeller=%d, BuyerSeller=%d, Elite=%d\n",
	   _traderCount, _pureBuyerCount, _pureSellerCount, _buysellCount, elite);
	printf("BuyerQuota=%d,  SellerQuota=%d\n", _buyerQuota, _sellerQuota);

	printf("InitExpPayoff=%+4.2f\n", _initExpPayoff );

	printf("H=%+4.2f, C=%+4.2f, D=%+4.2f, L=%+4.2f, R=%+6.3f, W=%+6.3f\n",
			_temptation, _bothCoop, _bothDefect, _sucker,
			_refusalPayoff, _wallflowerPayoff);

	printf("FsmStates=%d\n", fsmStates);
	printf("FsmMemory=%d\n\n", fsmMemory);
		
	if (_buysellCount < 0)
	{
		_buysellCount = 0;
	}

    ASSERT(_pureBuyerCount==0  || _pureBuyerCount >= elite >= 0);
    ASSERT(_pureSellerCount==0 || _pureSellerCount >= elite >= 0);
    ASSERT(_buysellCount==0    || _buysellCount >= elite >= 0);

	if (_pureBuyerCount > 0)
	{
		_pureBuyPopulation.Init();
		_pureBuyPopulation.SetMutationRate(mrate);
		_pureBuyPopulation.SetElite(elite);
		_pureBuyPopulation.SetSize(_pureBuyerCount);
	}

	if (_pureSellerCount > 0)
	{
		_pureSellPopulation.Init();
		_pureSellPopulation.SetMutationRate(mrate);
		_pureSellPopulation.SetElite(elite);
		_pureSellPopulation.SetSize(_pureSellerCount);
	}

	if (_buysellCount > 0)
	{
		_buysellPopulation.Init();
		_buysellPopulation.SetMutationRate(mrate);
		_buysellPopulation.SetElite(elite);
		_buysellPopulation.SetSize(_buysellCount);
	}
}

void
tngSimulation::InitGen(int g)
{
	int						i;
	tngTradeBot*			bot = 0;

	ASSERT(_traders.IsEmpty());
	ASSERT(_buyers.IsEmpty());
	ASSERT(_sellers.IsEmpty());

	_curGen = g;
	printf ("\nGeneration #%3d\n", _curGen+1);

	for (i = 0; i < _pureBuyerCount; i++)
	{
		bot = (tngTradeBot*) _pureBuyPopulation.GetPType(i);
		ASSERT_VALID(bot);

		_traders.Append(bot);
	}

	for (i = 0; i < _pureSellerCount; i++)
	{
		bot = (tngTradeBot*) _pureSellPopulation.GetPType(i);
		ASSERT_VALID(bot);

		_traders.Append(bot);
	}

	for (i = 0; i < _buysellCount; i++)
	{
		bot = (tngTradeBot*) _buysellPopulation.GetPType(i);
		ASSERT_VALID(bot);

		_traders.Append(bot);
	}

	for (i = 0; i < _pureBuyerCount; i++)
	{
		bot = (tngTradeBot*) _pureBuyPopulation.GetPType(i);
		ASSERT_VALID(bot);

		_buyers.Append(bot);

		bot->Init(this, _traders, _buyerQuota, 0);
	}

	for (i = 0; i < _pureSellerCount; i++)
	{
		bot = (tngTradeBot*) _pureSellPopulation.GetPType(i);
		ASSERT_VALID(bot);

		_sellers.Append(bot);

		bot->Init(this, _traders, 0, _sellerQuota);
	}

	for (i = 0; i < _buysellCount; i++)
	{
		bot = (tngTradeBot*) _buysellPopulation.GetPType(i);
		ASSERT_VALID(bot);

		_buyers.Append(bot);
		_sellers.Append(bot);

		bot->Init(this, _traders, _buyerQuota, _sellerQuota);
	}
}

void
tngSimulation::MatchTraders(void) 
{
	tngTradeBot* 		buyer;
	tngTradeBot* 		seller;
	tngTradeBot* 		trader;
	int					offers;

	for (_buyers.First(); buyer = (tngTradeBot*) _buyers.Next(); )
	{
		buyer->PrepareOffers();
	}
                   
// Tell all buyers to submit offers and all sellers to
// provisionally accept offers until the buyers stop making
// offers.
	while(1)
	{
		offers = 0;

// Tell all buyers to submit their offers...
		for (_buyers.First(); buyer = (tngTradeBot*) _buyers.Next(); )
		{
			offers += buyer->SubmitOffers();
		}

		if (offers == 0)
		{
			break;
		}

// Tell all sellers to provisionally accept their best offers...
		for (_sellers.First(); seller = (tngTradeBot*) _sellers.Next();)
		{
			seller->AcceptOffers();
		}
	}

	for (_traders.First(); trader = (tngTradeBot*) _traders.Next(); )
	{
		trader->WallflowerCheck();
	}
}

void
tngSimulation::Trade(int cycle)
{
	tngTradeBot* 		seller;

	// Set the current environment cycle to be this trade cycle
	_curEnv = cycle; 

	// Reset number of trades
	_trades = 0;

	// Tell all sellers to initiate a PD game with their respective
	// trade partners.
	for (_sellers.First(); seller = (tngTradeBot*) _sellers.Next(); )
	{
		seller->PlayPD();
	}

	if (_curGen >= (GMax() - _printGen)) 
	{
		if (_curEnv >= (IMax() - _printIt)) 
		{
			printf ("   Trade cycle #%4d:", _curEnv+1);
			printf ("   Total trades = %3d\n", _trades);
		}
	}
}

void
tngSimulation::AssessFitness(void)
{    
	tngTradeBot* 		bot;

	for (_traders.First(); bot = (tngTradeBot*) _traders.Next(); )
	{
		bot->CalcFitness();	

		if (_curGen == GMax()-1)
		{
			bot->Dump();
		}
	}
	
	_pureBuyPopulation.SortByFitness();
	_pureSellPopulation.SortByFitness();
	_buysellPopulation.SortByFitness();  
}

void
tngSimulation::Dump(void)
{  
// Determine the subpopulation and total population
// min, max, and average fitness scores.

// Total population values
   double   minFitness = +1000.0;
   double   avgFitness =     0.0;
   double   maxFitness = -1000.0;
   double   totFitness =     0.0;

// Subpopulation values
   double   minScore;
   double   avgScore;
   double   maxScore;
   double   totScore;

// Temporary variables
   int           i;
   tngTradeBot*  bot = 0;
   double        fitness = 0.0;

   if (_pureBuyerCount > 0)
   {
      minScore = +1000.0;
      maxScore = -1000.0;
      totScore =     0.0;

      for (i = 0; i < _pureBuyerCount; i++)
      {
         bot = (tngTradeBot*) _pureBuyPopulation.GetPType(i);
         ASSERT_VALID(bot);

         fitness = bot->CalcFitness() ;

         minScore = min(minScore, fitness);
         maxScore = max(maxScore, fitness);
         totScore += fitness;
      }
      avgScore = totScore/_pureBuyerCount ;

 printf("    PureBuyer scores...   min: %+5.4f max: %+5.4f avg: %+5.4f\n",
                minScore, maxScore, avgScore);

      totFitness += totScore;
      minFitness = min(minFitness, minScore);
      maxFitness = max(maxFitness, maxScore);
   }

   if (_pureSellerCount > 0)
   {
      minScore = +1000.0;
      maxScore = -1000.0;
      totScore = 0.0;

      for (i = 0; i < _pureSellerCount; i++)
      {
         bot = (tngTradeBot*) _pureSellPopulation.GetPType(i);
         ASSERT_VALID(bot);

         fitness = bot->CalcFitness() ;

         minScore = min(minScore, fitness);
         maxScore = max(maxScore, fitness);
         totScore += fitness;
      }
      avgScore = totScore/_pureSellerCount ;

 printf("    PureSeller scores...  min: %+5.4f max: %+5.4f avg: %+5.4f\n",
                minScore, maxScore, avgScore);

      totFitness += totScore;
      minFitness = min(minFitness, minScore);
      maxFitness = max(maxFitness, maxScore);
   }

   if (_buysellCount > 0)
   {
      minScore = +1000.0;
      maxScore = -1000.0;
      totScore = 0.0;

      for (i = 0; i < _buysellCount; i++)
      {
         bot = (tngTradeBot*) _buysellPopulation.GetPType(i);
         ASSERT_VALID(bot);

         fitness = bot->CalcFitness() ;

         minScore = min(minScore, fitness);
         maxScore = max(maxScore, fitness);
         totScore += fitness;
      }
      avgScore = totScore/_buysellCount ;

 printf("    BuyerSeller scores... min: %+5.4f max: %+5.4f avg: %+5.4f\n",
                minScore, maxScore, avgScore);

      totFitness += totScore;
      minFitness = min(minFitness, minScore);
      maxFitness = max(maxFitness, maxScore);
   }

   avgFitness = totFitness/_traderCount ;

 printf("    all fitness scores... min: %+5.4f max: %+5.4f avg: %+5.4f\n",
            minFitness, maxFitness, avgFitness);
}

void
tngSimulation::EvolveGen(void)
{    
	_traders.RemoveAll();
	_buyers.RemoveAll();
	_sellers.RemoveAll();
	                      
	_pureBuyPopulation.Breed();
	_pureSellPopulation.Breed();
	_buysellPopulation.Breed();  
}

float
tngSimulation::DilemmaPayoff(int mine, int his) 
//: Returns a payoff for the prisoner's dilemma
{       
	if (mine > 0)
	{
		if (his > 0)
		{
			return(float) _bothCoop;
		}
		else
		{            
			return(float) _sucker;
		}
	}
	else              
	{
		if (his > 0)
		{
			return(float) _temptation;
		}
		else
		{            
			return(float) _bothDefect;
		}
	}
}

void
tngSimulation::MediateTrade(tngTradeBot* seller, tngTradeBot* buyer)
//: Mediate a trade between two tradebots.
{ 
	int					sellerAction;
	int					buyerAction;
	                         
	float				sellerPayoff;
	float				buyerPayoff;
		                         
	ASSERT_VALID(seller);
	ASSERT_VALID(buyer);
	
	// Retrieve an action from each tradebot
	sellerAction = seller->GetAction(buyer);
	buyerAction = buyer->GetAction(seller);
	                                        
	// Determine a payoff for each action	                                        
	sellerPayoff = DilemmaPayoff(sellerAction, buyerAction);
	buyerPayoff = DilemmaPayoff(buyerAction, sellerAction);
                                         
	// Tell each tradebot what happened, allowing them to update their
	// respective information on each other.
	seller->Sell(buyer, sellerPayoff, buyerAction);
	buyer->Buy(seller, buyerPayoff, sellerAction);

	/*
	printf("\nTrade:(%s, %d, %5.4f) sold to (%s, %d, %5.4f)\n",
		seller->GetName(), sellerAction, sellerPayoff,
		buyer->GetName(), buyerAction, buyerPayoff);
	*/

	// Increment our trade count.		                                    
	++_trades;		
}
